#include <audio.h>
#include <math.h>
#include <windows.h>
#include <emu.h>

#include <portaudio.h>
#include <pa_ringbuffer.h>

#define SAMPLE_RATE 192000
#define FRAMES_PER_BUFFER 64
#define TIME_PER_SAMPLE 1.0f / (SAMPLE_RATE/1000.0f/1000.0f)
#define TIME_PER_AUDIO_TICK 1.0f / (1048576.0f/1000.0f/1000.0f)
#define LFSR_BASE_CLOCK 262144.0f

static audio_context ctx;

static float audio_time = 0;
static float lfsr_timer = 0;
static float lfsr_clock = LFSR_BASE_CLOCK;

const u8 square_sample_00[8] = {
    0x0,
    0xF,
    0xF,
    0xF,
    0xF,
    0xF,
    0xF,
    0xF
};

const u8 square_sample_01[8] = {
    0x0,
    0x0,
    0xF,
    0xF,
    0xF,
    0xF,
    0xF,
    0xF
};

const u8 square_sample_10[8] = {
    0x0,
    0x0,
    0x0,
    0x0,
    0xF,
    0xF,
    0xF,
    0xF
};

const u8 square_sample_11[8] = {
    0x0,
    0x0,
    0x0,
    0x0,
    0x0,
    0x0,
    0xF,
    0xF
};

const u8 *square_sample[4] = {
    square_sample_00,
    square_sample_01,
    square_sample_10,
    square_sample_11
};

static int audio_callback(const void* input_uffer, void *output_buffer,
                            unsigned long framesPerBuffer,
                            const PaStreamCallbackTimeInfo *time_info,
                            PaStreamCallbackFlags status_flags,
                            void *userData ) {
    float *out = (float *)output_buffer;
    float left = 0;
    float right = 0;
    for(int i = 0; i < framesPerBuffer; i++) {
        audio_time += TIME_PER_SAMPLE;
        for(;audio_time >= TIME_PER_AUDIO_TICK;audio_time -= TIME_PER_AUDIO_TICK) {
            ctx.sq1_period_timer++;
            ctx.sq2_period_timer++;
            ctx.ch3_period_timer++;

            if(ctx.sq1_period_timer >= 0x800) {
                ctx.sq1_period_timer = ctx.sq1_period_reset;
                ctx.sq1_sample = (ctx.sq1_sample + 1) % 8;
            }
            if(ctx.sq2_period_timer >= 0x800) {
                ctx.sq2_period_timer = ctx.sq2_period_reset;
                ctx.sq2_sample = (ctx.sq2_sample + 1) % 8;
            }
            if(ctx.ch3_period_timer >= 0x800) {
                ctx.ch3_period_timer = ctx.ch3_period_reset;
                ctx.ch3_sample = (ctx.ch3_sample + 1) % 32;
                if(ctx.ch3_sample & 0b1) {
                    ctx.ch3_last_sample = ctx.wave_ram[ctx.ch3_sample >> 1] & 0xF;
                } else {
                    ctx.ch3_last_sample = ctx.wave_ram[ctx.ch3_sample >> 1] >> 4;
                }
            }
        }
        lfsr_timer += TIME_PER_SAMPLE;
        for(;lfsr_timer >= lfsr_clock;lfsr_timer -= lfsr_clock) {
            if(lfsr_timer >= lfsr_clock && ctx.ch4_enable) {
                lfsr_timer = 0;
                u8 new = !((ctx.ch4_lfsr & 0b1) ^ ((ctx.ch4_lfsr >> 1) & 0b1)) & 0b1;
                ctx.ch4_lfsr |= (new << 15);
                if(ctx.ch4_lfsr_width) {
                    ctx.ch4_lfsr |= (new << 7);
                }
                ctx.ch4_lfsr = ctx.ch4_lfsr >> 1;
                //printf("lfsr: %02X, bit: %d\n", ctx.ch4_lfsr, new);
            }
        }
        if(ctx.audio_enabled){
            if(ctx.sq1_enable) {
                if(ctx.ch1_left) {
                    left += ((float)ctx.sq1_volume/15.0f) * (((float)(square_sample[ctx.sq1_duty][ctx.sq1_sample]) - 7.5f)/7.5f);
                }
                if(ctx.ch1_right) {
                    right += ((float)ctx.sq1_volume/15.0f) * (((float)square_sample[ctx.sq1_duty][ctx.sq1_sample] - 7.5f)/7.5f);
                }
            }

            if(ctx.sq2_enable) {
                if(ctx.ch2_left) {
                    left += ((float)ctx.sq2_volume/15.0f) * (((float)(square_sample[ctx.sq2_duty][ctx.sq2_sample]) - 7.5f)/7.5f);
                }
                if(ctx.ch2_right) {
                    right += ((float)ctx.sq2_volume/15.0f) * (((float)(square_sample[ctx.sq2_duty][ctx.sq2_sample]) - 7.5f)/7.5f);
                }
            }

            if(ctx.ch3_enable && ctx.ch3_volume != 00) {
                u8 shift = 0;
                if(ctx.ch3_volume == 0b10) {
                    shift = 1;
                }
                if(ctx.ch3_volume == 0b11) {
                    shift = 2;
                }
                if(ctx.ch3_left) {
                    left += (((float)(ctx.ch3_last_sample >> shift) - 7.5f)/7.5f);
                    //printf("left: %d\n", ctx.ch3_volume);
                }
                if(ctx.ch3_right) {
                    right += (((float)(ctx.ch3_last_sample >> shift) - 7.5f)/7.5f);
                }
            }

            if(ctx.ch4_enable) {
                if(ctx.ch4_left) {
                    left += ((float)ctx.ch4_volume/15.0f) * (ctx.ch4_lfsr & 0b1);
                    //printf("left: %d\n", ctx.ch3_volume);
                }
                if(ctx.ch4_right) {
                    right += ((float)ctx.ch4_volume/15.0f) * (ctx.ch4_lfsr & 0b1);
                }
            }
        }

        left *= (float)ctx.volume_left/7.0f;
        right *= (float)ctx.volume_right/7.0f;

        left /= 4;
        right /= 4;

        if(left > 1.0f) {
            printf("Uh Oh! %02X\n", ctx.volume_left);
        }
        *out++ = left;
        *out++ = right;
    }
    return paContinue;
}
static PaStream *stream;

void audio_sample();

void audio_init(){
    PaStreamParameters output_parameters;
    PaError err;

    ctx.audio_enabled = false;
    ctx.ch1_left = false;
    ctx.ch1_right = false;
    ctx.ch2_left = false;
    ctx.ch2_right = false;
    ctx.ch3_left = false;
    ctx.ch3_right = false;
    ctx.ch4_left = false;
    ctx.ch4_right = false;

    ctx.volume_left = 0x1;
    ctx.volume_right = 0x1;

    ctx.sq1_duty = 0x0;
    ctx.sq1_volume = 0x0;
    ctx.sq1_sample = 0x0;
    ctx.sq1_period_reset = 0x0;
    ctx.sq1_period_timer = 0x0;
    ctx.sq1_enable = false;
    ctx.sq1_len_enable = false;
    ctx.sq1_sweep_pace = 0x0;
    ctx.sq1_sweep_direction = false;
    ctx.sq1_initial_len = 0x0;
    ctx.sq1_len = 0x0;
    ctx.sq1_initial_volume = 0x0;
    ctx.sq1_env_direction = false;
    ctx.sq1_env_pace = 0x0;
    ctx.sq1_env_timer = 0x0;
    ctx.sq1_sweep_step = 0x0;

    ctx.sq2_duty = 0x0;
    ctx.sq2_volume = 0x0;
    ctx.sq2_sample = 0x0;
    ctx.sq2_period_reset = 0x0;
    ctx.sq2_period_timer = 0x0;
    ctx.sq2_enable = false;
    ctx.sq2_len_enable = false;
    ctx.sq2_initial_len = 0x0;
    ctx.sq2_len = 0x0;
    ctx.sq2_initial_volume = 0x0;
    ctx.sq2_env_direction = false;
    ctx.sq2_env_pace = 0x0;
    ctx.sq2_env_timer = 0x0;
    
    err = Pa_Initialize();
    if (err != paNoError) goto error;
    output_parameters.device = Pa_GetDefaultOutputDevice();
    if(output_parameters.device == paNoDevice) {
        fprintf(stderr, "No default audio device!\n");
        goto error;
    }
    output_parameters.channelCount = 2;
    output_parameters.sampleFormat = paFloat32;
    output_parameters.suggestedLatency = Pa_GetDeviceInfo(output_parameters.device)->defaultLowOutputLatency;
    output_parameters.hostApiSpecificStreamInfo = NULL;

    err = Pa_OpenStream(&stream,
                        NULL,
                        &output_parameters,
                        SAMPLE_RATE,
                        paFramesPerBufferUnspecified,
                        paNoFlag,
                        audio_callback,
                        &ctx);
    if(err != paNoError) goto error;
    err = Pa_StartStream(stream);
    if(err != paNoError) goto error;
    return;
error:
    Pa_Terminate();
    fprintf(stderr, "portaudio stream error\n\tError Number: %d\n\tError Message: %s\n", err, Pa_GetErrorText(err));
}

void sq1_sweep() {
    //frequency calculation
    int step = (ctx.sq1_sweep_period >> ctx.sq1_sweep_step);
    step = ctx.sq1_sweep_direction ? -step : step;
    ctx.sq1_calc_period = ctx.sq1_sweep_period + step;
    //overflow check
    if(ctx.sq1_calc_period > 0x7FFF) {
        ctx.sq1_enable = false;
    }
}

static int change = 1;
static u32 ticks = 0;

void audio_tick(){
    u32 prev_ticks = ticks;
    ticks++;
    if(!(ticks & 0b1)) {
        if(ctx.sq1_len_enable) {
            ctx.sq1_len++;
            if(ctx.sq1_len >= 64) {
                ctx.sq1_enable = false;
            }
        }

        if(ctx.sq2_len_enable) {
            ctx.sq2_len++;
            if(ctx.sq2_len >= 64) {
                ctx.sq2_enable = false;
            }
        }

        if(ctx.ch3_len_enable) {
            ctx.ch3_len++;
            if(ctx.ch3_len >= 64) {
                ctx.ch3_enable = false;
            }
        }

        if(ctx.ch4_len_enable) {
            ctx.ch4_len++;
            if(ctx.ch4_len >= 64) {
                ctx.ch4_enable = false;
            }
        }
    }

    if((prev_ticks & (1 << 3)) && !(ticks & (1 << 3))) {
        if(ctx.sq1_env_pace != 0){
            ctx.sq1_env_timer++;
            if(ctx.sq1_env_timer >= ctx.sq1_env_pace) {
                ctx.sq1_env_timer = 0;
                if((ctx.sq1_env_direction && ctx.sq1_volume != 15) || (!ctx.sq1_env_direction && ctx.sq1_volume != 0)){
                    ctx.sq1_volume += ctx.sq1_env_direction ? 1 : -1;
                    if(ctx.sq1_volume < 0)
                        ctx.sq1_volume = 0;
                    if(ctx.sq1_volume > 15)
                        ctx.sq1_volume = 15;
                }
            }
        }

        if(ctx.sq2_env_pace != 0){
            ctx.sq2_env_timer++;
            if(ctx.sq2_env_timer >= ctx.sq2_env_pace) {
                ctx.sq2_env_timer = 0;
                if((ctx.sq2_env_direction && ctx.sq2_volume != 15) || (!ctx.sq2_env_direction && ctx.sq2_volume != 0)){
                    ctx.sq2_volume += ctx.sq2_env_direction ? 1 : -1;
                    if(ctx.sq2_volume < 0)
                        ctx.sq2_volume = 0;
                    if(ctx.sq2_volume > 15)
                        ctx.sq2_volume = 15;
                }
                //printf("sq2 vol: %01X\n", ctx.sq2_volume);
            }
        }

        if(ctx.ch4_env_pace != 0){
            ctx.ch4_env_timer++;
            if(ctx.ch4_env_timer >= ctx.ch4_env_pace) {
                ctx.ch4_env_timer = 0;
                if((ctx.ch4_env_direction && ctx.ch4_volume != 15) || (!ctx.ch4_env_direction && ctx.ch4_volume != 0)){
                    ctx.ch4_volume += ctx.ch4_env_direction ? 1 : -1;
                    if(ctx.ch4_volume < 0)
                        ctx.ch4_volume = 0;
                    if(ctx.ch4_volume > 15)
                        ctx.ch4_volume = 15;
                }
                //printf("sq2 vol: %01X\n", ctx.sq2_volume);
            }
        }

    }

    if((prev_ticks & (1 << 2)) && !(ticks & (1 << 2))) {
        ctx.sq1_sweep_timer++;
        if(ctx.sq1_sweep_timer >= ctx.sq1_sweep_pace) {
            ctx.sq1_sweep_timer = 0;
            if(ctx.sq1_enable && ctx.sq1_sweep_pace){
                sq1_sweep();
                if(ctx.sq1_calc_period <= 0x7FFF && ctx.sq1_sweep_step != 0) {
                    ctx.sq1_sweep_period = ctx.sq1_calc_period;
                    ctx.sq1_period_reset = ctx.sq1_calc_period;
                }
                sq1_sweep();
            }
        }
    }
}

void enable_square1() {
    ctx.sq1_enable = true;
    if(ctx.sq1_len >= 64) {
        ctx.sq1_len = ctx.sq1_initial_len;
    }
    ctx.sq1_volume = ctx.sq1_initial_volume;
    ctx.sq1_env_timer = 0;
    ctx.sq1_sweep_period = ctx.sq1_period_reset;
    ctx.sq1_sweep_timer = 0;
    ctx.sq1_env_direction = ctx.sq1_env_direction_buffer;
    ctx.sq1_env_pace = ctx.sq1_env_pace_buffer;
    ctx.sq1_sweep_enabled = (ctx.sq1_sweep_pace || ctx.sq1_sweep_step);
    if(ctx.sq1_sweep_step) {
        sq1_sweep();
    }
}

void enable_square2() {
    ctx.sq2_enable = true;
    if(ctx.sq2_len >= 64) {
        ctx.sq2_len = ctx.sq2_initial_len;
    }
    ctx.sq2_volume = ctx.sq2_initial_volume;
    ctx.sq2_env_timer = 0;
    ctx.sq2_sample = 0;
    ctx.sq2_period_timer = ctx.sq2_period_reset;
    ctx.sq2_env_direction = ctx.sq2_env_direction_buffer;
    ctx.sq2_env_pace = ctx.sq2_env_pace_buffer;
}

void enable_wave() {
    ctx.ch3_enable = true;
    if(ctx.ch3_len >= 64) {
        ctx.ch3_len = ctx.ch3_initial_len;
    }
    ctx.ch3_sample = 0;
    ctx.ch3_volume = ctx.ch3_initial_volume;
    ctx.ch3_period_timer = ctx.ch3_period_reset;
}

void enable_noise() {
    ctx.ch4_enable = true;
    if(ctx.ch4_len >= 64) {
        ctx.ch4_len = ctx.ch4_initial_len;
    }
    ctx.ch4_env_timer = 0;
    ctx.ch4_env_direction = ctx.ch4_env_direction_buffer;
    ctx.ch4_env_pace = ctx.ch4_env_pace_buffer;
    ctx.ch4_volume = ctx.ch4_initial_volume;
    ctx.ch4_lfsr = 0;
}

u8 audio_read(u16 address) {
    return 0x00;
}

void audio_write(u16 address, u8 value){
    if(address == 0xFF26) {
        ctx.audio_enabled = value & 0x80;
    }
    if(address == 0xFF25) {
        ctx.ch4_left = value & 0b10000000;
        ctx.ch3_left = value & 0b01000000;
        ctx.ch2_left = value & 0b00100000;
        ctx.ch1_left = value & 0b00010000;

        ctx.ch4_right = value & 0b00001000;
        ctx.ch3_right = value & 0b00000100;
        ctx.ch2_right = value & 0b00000010;
        ctx.ch1_right = value & 0b00000001;
    }

    if(address == 0xFF24) {
        ctx.volume_left = (value >> 4) & 0b111;
        if(ctx.volume_left == 0) ctx.volume_left = 1;
        ctx.volume_right = value & 0b111;
        if(ctx.volume_right == 0) ctx.volume_right = 1;
    }

    if(address == 0xFF10) {
        ctx.sq1_sweep_pace = (value >> 4) & 0b111;
        ctx.sq1_sweep_direction = value & 0x08;
        ctx.sq1_sweep_step = value & 0b111;
    }
    if(address == 0xFF11) {
        ctx.sq1_duty = value >> 6;
        ctx.sq1_initial_len = value & 0x3F;
    }
    if(address == 0xFF12) {
        ctx.sq1_initial_volume = (value >> 4) & 0x0F;
        ctx.sq1_env_direction_buffer = (value & 0x8) == 0x80;
        ctx.sq1_env_pace_buffer = value & 0b111;
    }
    if(address == 0xFF13) {
        ctx.sq1_period_reset = (ctx.sq1_period_reset & 0xF00) | value;
    }
    if(address == 0xFF14) {
        ctx.sq1_period_reset = (ctx.sq1_period_reset & 0x0FF) | ((value & 0b111) << 8);
        ctx.sq1_len_enable = (value & 0x40) == 0x40;
        if(value & 0x80) {
            enable_square1();
        }
    }


    if(address == 0xFF16) {
        ctx.sq2_duty = value >> 6;
        ctx.sq2_initial_len = value & 0x3F;
    }
    if(address == 0xFF17) {
        ctx.sq2_initial_volume = (value >> 4) & 0x0F;
        ctx.sq2_env_direction_buffer = (value & 0x8) == 0x80;
        ctx.sq2_env_pace_buffer = value & 0b111;
        if(ctx.sq2_env_direction == 0 && ctx.sq2_initial_volume == 0) {
            ctx.sq2_enable = false;
        }
    }
    if(address == 0xFF18) {
        u16 prev_period = ctx.sq2_period_reset;
        ctx.sq2_period_reset = (ctx.sq2_period_reset & 0xF00) | value;
        //printf("period: %03X, old_period: %03X\n", ctx.sq2_period_reset, prev_period);
    }
    if(address == 0xFF19) {
        ctx.sq2_period_reset = (ctx.sq2_period_reset & 0x0FF) | ((value & 0b111) << 8);
        ctx.sq2_len_enable = (value & 0x40) == 0x40;
        if((value & 0x80) == 0x80) {
            enable_square2();
        }
    }

    if(address == 0xFF1A) {
        //ctx.ch3_enable = (value & 0x80) == 0x80;
    }
    
    if(address == 0xFF1B) {
        ctx.ch3_initial_len = value;
    }

    if(address == 0xFF1C) {
        ctx.ch3_initial_volume = (value >> 5) & 0b11;
    }

    if(address == 0xFF1D) {
        ctx.ch3_period_reset = (ctx.ch3_period_reset & 0xF00) | value;
    }

    if(address == 0xFF1E) {
        ctx.ch3_period_reset = (ctx.ch3_period_reset & 0x0FF) | ((value & 0b111) << 8);
        ctx.ch3_len_enable = (value & 0x40) == 0x40;
        if(value & 0x80) {
            enable_wave();
        }
    }

    if(address == 0xFF20) {
        ctx.ch4_initial_len = value & 0b111111;
    }

    if(address == 0xFF21) {
        ctx.ch4_initial_volume = (value >> 4) & 0x0F;
        ctx.ch4_env_direction_buffer = (value & 0x8) == 0x80;
        ctx.ch4_env_pace_buffer = value & 0b111;
    }

    if(address == 0xFF22) {
        ctx.ch4_clock_shift = (value >> 4);
        ctx.ch4_lfsr_width = (value & 0x08);
        ctx.ch4_clock_divider = value & 0b111;
        float div = (ctx.ch4_clock_divider == 0 ? 0.5f : ctx.ch4_clock_divider);
        float lfsr_rate = (LFSR_BASE_CLOCK) / div * (1 << ctx.ch4_clock_shift);
        lfsr_clock = 1.0f / (lfsr_rate/1000.0f/1000.0f);
    }

    if(address == 0xFF23) {
        ctx.ch4_len_enable = (value & 0x40) == 0x40;
        if((value & 0x80) == 0x80) {
            enable_noise();
        }
    }

    if(BETWEEN(address, 0xFF30, 0xFF3F)) {
        ctx.wave_ram[address - 0xFF30] = value;
    }
}